@@ -30,14 +30,17 @@ require 'CTioga/tioga_primitives'
 # The debugging facility
 require 'CTioga/debug'
 require 'CTioga/log'
+require 'CTioga/utils'
+require 'CTioga/layout'
+
 
 # The backends
 require 'CTioga/backends'
 
 # Support for themes
 require 'CTioga/themes'
-
-
+require 'CTioga/axes'
+require 'CTioga/structures'
 
 
 # for interpreting the CTIOGA environment variable
@@ -94,13 +97,12 @@ module CTioga
 A class to test visually the effects of different stuff
 EOD
 
-    
-
     include SciYAG
     include Tioga
 
     include Debug
     include Log
+    include Utils               # For the safe_float function
 
     # For dimension conversion and TeX quoting.
     include Tioga::Utils
@@ -110,6 +112,7 @@ EOD
 
     # Support for Themes:
     include Themes
+    include Axes
 
 
     # these are basically the attributes which are modified directly
@@ -129,6 +132,12 @@ EOD
     # The current object
     attr_accessor :current_object
 
+    # Whether we are making a PNG file
+    attr_accessor :png
+
+    # The PNG's size:
+    attr_accessor :png_size
+
     def initialize
       # The first thing to do is to setup logging, as you really
       # want to be able to communicate with users, don't you ?
@@ -138,7 +147,6 @@ EOD
       @args = []                # Holding the command-line
       @parser = OptionParser.new
 
-
       initialize_themes
 
       # Whether the plots should be interpolated.
@@ -153,18 +161,12 @@ EOD
 
       # Initialize the debugging options:
 #      init_debug
-
-      # The multiplication factor for the plots.
-      # Nothing happens by default.
-      @x_factor = nil
-      @y_factor = nil
-
-      # if the axes are log scale
-      @x_log = false
-      @y_log = false
+      init_axes
 
       # now, the elements for the structure of the plot:
       @root_object = SubPlot.new
+      # We start with a very simple layout
+      SimpleLayout.new(@root_object)
       
       # the current object is the one to which we'll add plot elements.
       @current_object = @root_object
@@ -174,7 +176,7 @@ EOD
       @legend = nil
 
       # general purpose variables:
-      @cleanup = false
+      @cleanup = true           # Now, cleaning is done by default !
       @viewer = false
 
       # Some options we could set on the command line
@@ -193,6 +195,9 @@ EOD
 
       # The frame sides for the setup_real_size function
       @frame_sides = [0.1,0.9,0.9,0.1]
+      # Override this, as the layout scheme makes it more or less
+      # obsolete
+      @frame_sides = [0,1,1,0]
 
       # If set, we create separate legend informations:
       @separate_legend = false
@@ -204,7 +209,7 @@ EOD
       @separate_legend_height = 8
 
       # Whether to mark the command-line in the PDF file
-      @mark = false             # better be off, kind of painful sometimes
+      @mark = true              # On by default, really useful !!!
 
       # The LaTeX preamble:
       @preamble = ""
@@ -212,11 +217,56 @@ EOD
       # The last curve used:
       @last_curve = nil
 
+      # Specifications for the --next stuff: an array, the first
+      # element is the class to be created for children and the
+      # rest are arguments to be added at the beginning.
+      @next_specs = [SubPlot, :subplot]
+
+      # A block to be run on both the old and new object
+      # when --next is encountered.
+      @next_block = nil
+
+      # Whether we automatically start a --next stuff on each spec
+      # or even on each dataset
+      @auto_next = false
+
+      # We don't make any PNG file:
+      @png = false
+
+      # The standard PNG density...
+      @png_oversampling = 2
+
+      # The default padding;
+      @default_padding = []
+      4.times do 
+        @default_padding << Dimension.new(0.05)
+      end
+
+      # And we use it :
+      use_default_padding
+      
       prepare_option_parser
     end
 
+    # Sets the given's object padding to the current default:
+    def use_default_padding(obj = nil)
+      obj = current_object unless obj
+      4.times do |i|
+        obj.layout_preferences.padding[i] = @default_padding[i].dup
+      end
+      debug "Setting padding for #{identify(obj)} to " +
+        obj.layout_preferences.padding.inspect
+    end
 
 
+    # Adds the given object to the current's stack and set it
+    # as the current object, so that it will receive further
+    # children
+    def enter_child_object(object)
+      @current_object.add_elem(object)
+      self.current_object = object
+    end
+
 
 
     # This function reads a configuration file and executes it's
@@ -256,6 +306,13 @@ EOD
       return nil
     end
 
+    # Sets the frame of the root object. This is absolutely
+    # necessary for layout computation. Failure to do so will
+    # result in failed plots.
+    def set_root_frame(*frames)
+        @root_object.root_frame = frames
+    end
+
     # sets up the FigureMaker object to use for real size. Uses the
     # @real_size instance variable as a source for the size
     def setup_real_size(t)
@@ -267,6 +324,10 @@ EOD
       t.def_enter_page_function {
         t.page_setup(*sizes)
         t.set_frame_sides(*@frame_sides) 
+        set_root_frame(sizes[0] * @frame_sides[0], 
+                       sizes[0] * @frame_sides[1],
+                       sizes[1] * @frame_sides[2], 
+                       sizes[1] * @frame_sides[3])
       }
 
       # Setting label and title scale to 1
@@ -283,33 +344,6 @@ EOD
     end
 
 
-    # Sets the value of a string hosted by @current_object
-    def current_set_string(name)
-      return proc {|s|
-        if s
-          @current_object.send(name,s)
-        else
-          @current_object.send(name,nil)
-        end
-      }
-    end
-
-    def evaluate_float(s)
-      return Float(eval(s))
-    end
-
-    def bool_arg(str)
-      if str =~ TRUE_RE || str.nil?
-        true
-      else
-        false
-      end
-    end
-
-    def cleanup=(arg)
-      @cleanup = bool_arg(arg)
-    end
-
     # Push a function call onto the stack of the current element
     def add_elem_funcall(sym, *args)
       @current_object.add_funcall(TiogaFuncall.new(sym, *args))
@@ -323,7 +357,7 @@ EOD
 
     # Forwards the boundary settings to the current object.
     def set_bounds(which, val)
-      method = ("bound_" + which + "=").to_sym
+      method = ("bound_#{which}=").to_sym
       @current_object.send(method, val)
     end
 
@@ -335,13 +369,26 @@ EOD
       set_bounds(b,last)
     end
 
-    # Returns an array suitable for use with set_frame_margins based on
-    # some text. The array is duplicated until it has at least four elements.
-    def margins_from_text(txt)
-      ary = txt.split(/\s*,\s*/).map {|s| s.to_f}
-      while ary.length < 4
-        ary += ary
+    # Splits a text into four components, expanding if necessary:
+    # * if there is only one element, all four become this one
+    # * if there are two: w,h, it becomes w,w,h,h
+    # * if there are three, the last element is duplicated.
+    def expand_sides(txt)
+      ary = txt.split(/\s*,\s*/)
+      case ary.size
+      when 1
+        return [ary[0], ary[0], ary[0], ary[0]]
+      when 2
+        return [ary[0], ary[0], ary[1], ary[1]]
+      when 3
+        return [ary[0], ary[1], ary[2], ary[2]]
+      else
+        return ary
       end
+    end
+
+    def margins_from_text(txt)
+      ary = expand_sides(txt).map {|s| s.to_f}
       return ary
     end
 
@@ -351,38 +398,6 @@ EOD
       @frame_sides = [val[0], 1.0 - val[1], 1.0 - val[3], val[2]]
     end
 
-    # A hash to help choose which elements we need to modify in the
-    # legends spec hash.
-    Legend_Specs_Sides = {
-      'right' => ['plot_right_margin', 'legend_left_margin'],
-      'left' => ['plot_left_margin', 'legend_right_margin'],
-      'top' => ['plot_top_margin', 'legend_bottom_margin'],
-      'bottom' => ['plot_bottom_margin', 'legend_top_margin'],
-    }
-
-    # This function sets up the bounds for plots/legend based on
-    # a _side_ and _size_ attribute. Without any arguments,
-    # just return an empty
-    def prepare_legend_specs(side = false, size = 0.0, delta = 0.0 )
-      specs = {}
-      # Resetting specs to 0
-      for a in %w(top bottom left right)
-        specs["legend_#{a}_margin"] = 0.0
-        specs["plot_#{a}_margin"] = 0.0
-      end
-      if side
-        if Legend_Specs_Sides.key?(side.downcase)
-          a,b = *Legend_Specs_Sides[side.downcase]
-          specs[a] = size
-          specs[b] = 1 - size + delta
-        else
-          error "Legend side #{side} unknown"
-        end
-      end
-      return specs
-    end
-
-
     # A function that transforms an inset specification into
     # margins.
     def inset_margins(spec)
@@ -393,7 +408,6 @@ EOD
       return margins
     end
 
-
     # This function prepares the parser by giving it the arguments and
     # some small help text going along with them. This function will
     # unfortunately get really large with time, but there's nothing
@@ -412,27 +426,6 @@ EOD
         @preamble += "#{w}\n"
       end
 
-
-      @parser.on("--[no-]histogram",
-                 "All the next curves will be drawn", "as histograms") do |w|
-        @histogram = w
-      end
-
-
-
-#       # Sets:
-#       for arg in [[@colors, "color"],
-#                   [@markers, "marker"],
-#                   [@markers_colors, "marker-color"],
-#                   [@linestyle, "line-style"]]
-#         @parser.on("--#{arg[1]}-set SET", 
-#                    "Choose the current set for #{arg[1]} ",
-#                    "(#{arg[0].available_sets.join(' ')})",
-#                    moving_array_set(arg[0]))
-#       end
-
-
-
       @parser.separator "\nGlobal look:"
 
       @parser.on("--background COLOR",
@@ -472,48 +465,80 @@ EOD
 
       @parser.on("--rescale FACTOR",
                  "Scales everything by a given factor",
-                 "(useful for subplots)") do
-        |fact|
-        add_elem_funcall(:rescale, fact.to_f)
+                 "(useful for subplots)") do |fact|
+        current_object.rescale = safe_float(fact)
       end
 
-
-      @parser.separator "\nVarious axis stuff:"
-      @parser.on("--xfact FACT",
-                 "Multiply all x values by FACT") do |f|
-        @x_factor = evaluate_float(f)
-      end
-      @parser.on("--yfact FACT",
-                 "Multiply all y values by FACT") do |f|
-        @y_factor = evaluate_float(f)
-      end
-      @parser.on("--[no-]xlog",
-                 "Uses logarithmic scale for X axis") do |v|
-        @x_log = v
-        add_elem_funcall(:xaxis_log_values=, v)
-      end
-      @parser.on("--[no-]ylog",
-                 "Uses logarithmic scale for Y axis") do |v|
-        @y_log = v
-        add_elem_funcall(:yaxis_log_values=, v)
-      end
-      @parser.on("--comma",
-                 "Uses a comma for the decimal separator") do 
-        @decimal_separator = ','
-      end
-      @parser.on("--decimal SEP",
-                 "Uses SEP for the decimal separator") do |s|
-        @decimal_separator = s
+      @parser.on("--padding PADDING",
+                 "Changes the padding for the current object",
+                 "and subsequent ones") do |pad|
+        ary = expand_sides(pad)
+        @default_padding = ary.map {|dim| Dimension.new(dim)}
+        use_default_padding
       end
 
+      axes_options(@parser)
 
       @parser.separator "\nSubfigures, subplots, regions..."
-      @parser.on("--subplot",
-                 "Start a subplot") do
-        plot = SubPlot.new(:subplot, current_object)
-        @current_object.add_elem(plot)
-        self.current_object = plot
+#       @parser.on("--subplot",
+#                  "Start a subplot") do
+#         plot = SubPlot.new(:subplot, current_object)
+#         @current_object.add_elem(plot)
+#         self.current_object = plot
+#       end
+
+
+      @parser.on("--y2",
+                 "Switch to an alternative Y axis") do
+        begin
+          current_object.disable_axis(:y)
+        rescue
+          # Purely ignore errors
+          warn "Something should have happened here "+
+            "that did not happen properly"
+        end
+        plot = SharedAxisPlot.new(:y, current_object)
+        # The label will be on the right
+        plot.ylabel.side = RIGHT
+        plot.edges.yaxis.loc = RIGHT
+        plot.edges.set_edges(AXIS_HIDDEN, :left,:top,:bottom)
+        current_object.edges.set_edges(AXIS_LINE_ONLY, :right)
+        # A simple layout will control the plot
+        layout = FixedLayout.new(plot)        
+        # And this layout should report its info to
+        # the current one.
+        #
+        # Note that we add the child's *layout* to the
+        # objects layout, not the child itself.
+        current_object.layout.add_child(layout)
+        enter_child_object(plot)
       end
+
+      @parser.on("--x2",
+                 "Switch to an alternative X axis") do
+        begin
+          current_object.disable_axis(:y)
+        rescue
+          # Purely ignore errors
+          warn "Something should have happened here "+
+            "that did not happen properly"
+        end
+        plot = SharedAxisPlot.new(:x, current_object)
+        # The label will be on the right
+        plot.xlabel.side = TOP
+        plot.edges.xaxis.loc = TOP
+        plot.edges.set_edges(AXIS_HIDDEN, :left,:right,:bottom)
+        current_object.edges.set_edges(AXIS_LINE_ONLY, :top)
+        layout = FixedLayout.new(plot)        
+        # And this layout should report its info to
+        # the current one.
+        #
+        # Note that we add the child's *layout* to the
+        # objects layout, not the child itself.
+        current_object.layout.add_child(layout)
+        enter_child_object(plot)
+      end
+
       @parser.on("--inset SPECS",
                  "Creates an inset with the given",
                  "specifications: " +
@@ -523,20 +548,165 @@ EOD
                  "All numbers " +
                  "are between 0 and 1" ) do |a|
         plot = SubPlot.new(:subplot, current_object)
+        plot.show_legend = false
         @current_object.add_elem(plot)
         self.current_object = plot
         margins = inset_margins(a)
         debug "inset margins: #{margins.inspect}"
-        current_object.frame_margins = margins
+        # With force_position, there is no need for a layout.
+        current_object.force_position = margins
+        # We redirect all legends for the inset to the parent.
+        current_object.accept_legend = false
       end
-      @parser.on("--subframes FRAMES",
-                 "Setup the frames relative to the parent", 
-                 "for the current " +
-                 "subplot",
-                 "(distance from left,right,top,bottom)") do |v|
-        current_object.frame_margins = margins_from_text(v)
+
+      @parser.on("--zoom-inset SPECS",
+                 "Creates an inset as with --inset " +
+                 "and copies all the ",
+                 "elements in the current plot there") do |a|
+        plot = SubPlot.new(:subplot, current_object)
+        # We redirect all legends for the inset to the parent.
+        plot.accept_legend = false
+        # Temporarily disable legends.
+        plot.disable_legend = true
+        for obj in @current_object.elements
+          plot.add_elem(obj.dup)
+        end
+        for obj in @current_object.funcalls
+          plot.add_funcall(obj.dup)
+        end
+        plot.disable_legend = false
+        # Better add it afterwards...
+        @current_object.add_elem(plot)
+        self.current_object = plot
+        margins = inset_margins(a)
+        debug "zoom margins: #{margins.inspect}"
+        current_object.force_position = margins
       end
 
+      @parser.on("--disable-legends",
+                 "Disable the display of legends for the current " +
+                 "subplot/subfigure") do
+        current_object.disable_legend = true
+      end
+
+      @parser.on("--enable-legends",
+                 "Reverts --disable-legends") do
+        current_object.disable_legend = false
+      end
+
+      # Black magic is starting here !
+      @parser.on("--grid SPEC",
+                 "Creates a grid. SPEC is column|row=nb") do |s|
+        # get the spec:
+        s =~ /(\w+)=(\d+)/
+        which = "#{$1}s=".to_sym
+        number = $2.to_i
+        if which == :columns= or which == :rows=
+          current_object.layout = current_object.layout.
+            convert_layout(GridLayout)
+          current_object.edges.disable_axis_display(:x, AXIS_HIDDEN)
+          current_object.edges.disable_axis_display(:y, AXIS_HIDDEN)
+
+          # We rescale the padding by a factor of 1/number, so it looks
+          # reasonable in the end.
+          @default_padding.each do |dim|
+            dim.scale!(1.0/number)
+          end
+
+          # We use the default padding for the chidren
+          use_default_padding
+
+          current_object.xlabel.label = nil
+          current_object.ylabel.label = nil
+          # We make two columns, just for fun !
+          current_object.layout.send(which, number)
+          plot = SubPlot.new(:subplot, current_object)
+          plot.accept_legend = false
+          @current_object.layout.add_child(GridItemLayout.new(plot))
+          @current_object.add_elem(plot)
+          self.current_object = plot
+          @next_spec = [SubPlot, :subplot]
+          @next_block = proc do |old, new|
+            new.accept_legend = false # By default, forward to the parent.
+            use_default_padding(new)
+          end
+        else
+          warn "Unrecognized spec #$1"
+        end
+      end
+
+      # Black magic is starting here !
+      @parser.on("--col",
+                 "Starts a column of graphes with shared X axis") do
+        # We convert the current object to a GridLayout. If
+        # there are already plots, that really won't look good.
+        # You've been warned !!!
+        current_object.layout = current_object.layout.
+          convert_layout(GridLayout)
+#        current_object.edges.disable
+        current_object.edges.disable_axis_display(:x, AXIS_HIDDEN)
+        current_object.edges.disable_axis_display(:y, AXIS_HIDDEN)
+        current_object.layout_preferences.padding[2] = Dimension.new(0.0)
+        current_object.layout_preferences.padding[3] = Dimension.new(0.0)
+        # We make two columns, just for fun !
+        current_object.layout.columns = 1
+        plot = SharedAxisPlot.new(:y, current_object)
+        @next_spec = [SharedAxisPlot, :y]
+        @current_object.layout.add_child(GridItemLayout.new(plot))
+        @current_object.add_elem(plot)
+        plot.xlabel.label = current_object.xlabel.label
+        current_object.xlabel.label = nil
+        plot.title.label = current_object.title.label
+        current_object.title.label = nil
+        self.current_object = plot
+
+        # We cancel vertical padding by default:
+        @default_padding[2] = Dimension.new(0.0)
+        @default_padding[3] = Dimension.new(0.0)
+
+        use_default_padding(plot)
+        # Forwards the x label to the child, and cancels it here.
+        # The code to be run on the old and new tings
+        @next_block = proc do |old, new|
+          old.edges.disable_axis_display(:x)
+          new.xlabel.label = old.xlabel.label if old.xlabel.label
+          old.xlabel.label = nil
+          new.ylabel.label = old.ylabel.label if old.ylabel.label
+          use_default_padding(new)
+        end
+      end
+
+      @parser.on("--next",
+                 "Start the next object") do
+        next_object
+      end
+
+      @parser.on("--[no-]auto-next",
+                 "Automatically start a --next graph",
+                 "for every element on the command-line") do |v|
+        if v
+          @auto_next = :element
+        else
+          @auto_next = false
+        end
+      end
+
+      @parser.on("--auto-next-expanded",
+                 "Automatically start a --next graph",
+                 "for every element on the command-line",
+                 "after data set expansion !") do
+        @auto_next = :dataset
+      end
+
+
+#       @parser.on("--subframes FRAMES",
+#                  "Setup the frames relative to the parent", 
+#                  "for the current " +
+#                  "subplot",
+#                  "(distance from left,right,top,bottom)") do |v|
+#         current_object.frame_margins = margins_from_text(v)
+#       end
+
       @parser.on("--region",
                  "Start a colored region") do
         plot = Region.new(current_object)
@@ -544,6 +714,66 @@ EOD
         self.current_object = plot
       end
 
+      @parser.on("--region-color COLOR",
+                 "The color of the region") do |v|
+        c = CTioga.get_tioga_color(v)
+        begin
+          current_object.region_color = c
+        rescue
+          error "The current container isn't a --region !"
+        end
+      end
+
+      @parser.on("--region-transparency TRANS",
+                 "The transparency of the region") do |v|
+        c = v.to_f
+        begin
+          current_object.region_transparency = c
+        rescue
+          error "The current container isn't a --region !"
+        end
+      end
+
+      @parser.on("--region-dont-display",
+                 "No curve until the next --end will actually be ",
+                 "displayed: they are just used to delimit the ",
+                 "region used for the fills") do
+        begin
+          current_object.dont_display = true
+        rescue
+          error "The current container isn't a --region !"
+        end
+      end
+
+      @parser.on("--region-debug",
+                 "Setup the fills for the curves inside the",
+                 "region to help understanding what happens") do
+        begin
+          current_object.region_debug = true
+        rescue
+          error "The current container isn't a --region !"
+        end
+      end
+
+      @parser.on("--region-invert-rule",
+                 "Inverts the rule for choosing the filled region") do
+        begin
+          current_object.invert_rule = true
+        rescue
+          error "The current container isn't a --region !"
+        end
+      end
+
+      @parser.on("--region-fill-twice",
+                 "Fills the region twice, choosing opposite",
+                 "rules for the boundaries.") do
+        begin
+          current_object.fill_twice = true
+        rescue
+          error "The current container isn't a --region !"
+        end
+      end
+
       @parser.on("--end",
                  "Ends the last subplot or region") do 
         if current_object.parent
@@ -563,30 +793,21 @@ EOD
         @real_size = spec
       end
       @parser.on("--frame-margins VAL",
-                 "The proportion of space to leave on", "each side " +
-                 "for the text display" ) do |val|
+                 "The proportion of space to leave on", 
+                 "each side for the text display" ) do |val|
         set_frame_margins(margins_from_text(val))
       end
 
 
       @parser.separator "\nLegends and texts:"
-
-      @parser.on("-x","--xlabel [LABEL]",
-                 "Label of the x axis",
-                 current_set_string(:x_label=))
-      @parser.on("-y","--ylabel [LABEL]",
-                 "Label of the y axis",
-                 current_set_string(:y_label=))
-      @parser.on('-t', "--title [TITLE]",
-                 "Sets the title of the plot",
-                 current_set_string(:title=))
       @parser.on("-l","--[no-]legend [LEGEND]",
                  "Sets the legend of the next curve.") do |l|
         override_style.legend = l
       end
 
       @parser.on("--[no-]auto-legend", 
-                 "Whether to automatically add legends","to curves") do |l|
+                 "Whether to automatically add legends",
+                 "to curves") do |l|
         @autolegends = l
       end
 
@@ -596,7 +817,7 @@ EOD
       end
 
       ## Default TeX fontsize
-      @parser.on("--fontsize [NB]",
+      @parser.on("--fontsize NB",
                  "Default TeX fontsize in points") do |size|
         @tex_fontsize = size
       end
@@ -622,23 +843,22 @@ EOD
                  "Sets the legend position and its size.",
                  "SPEC looks like pos,size[,delta]") do |a|
         w,s,d = a.split /\s*,\s*/
+        if d
+          warn "The delta specification (#{d}) for #{a} is ignored"
+        end
         debug "Legend position #{w.inspect} #{s.inspect}"
-        specs = prepare_legend_specs(w, Float(s), d.to_f)
-        debug "Specs: #{specs.inspect}"
-        current_object.legend_specs = specs
+        current_object.layout_preferences.legend_position = w.to_sym
+        current_object.layout_preferences.legend_size = Dimension.new(s)
       end
 
       @parser.on("--legend-inside SPEC", 
                  "Makes a legend inside the plot.",
                  "See --inset for the format of SPEC") do |a|
         margins = inset_margins(a)
-        t = %w(legend_left_margin legend_right_margin 
-          legend_top_margin legend_bottom_margin).zip(margins)
-        specs = prepare_legend_specs
-        t.inject(specs) { |s,o| s[o[0]] = o[1]; s}
-        
+        specs = margins.to_frame("legend_%s_margin")
         debug "Specs: #{specs.inspect}"
-        current_object.legend_specs = specs
+        current_object.layout_preferences.legend_position = :inside
+        current_object.layout_preferences.legend_spec = specs
       end
 
 
@@ -647,7 +867,7 @@ EOD
                  "Draw a graphic primitive") do |spec|
         add_elems_to_current(TiogaPrimitiveMaker.parse_spec(spec, self))
       end
-      @parser.on("-d","--draw-help",
+      @parser.on("--draw-help",
                  "Display what is known about graphic", 
                  "primitives and exit") do
         puts
@@ -662,7 +882,7 @@ EOD
       @parser.separator "\nHousekeeping and miscellaneous options:"
       @parser.on("--xpdf", 
                  "Runs xpdf to show the plot obtained") do 
-        @viewer = "xpdf"
+        @viewer = "xpdf -z page" # With the zoom set to full page
       end
 
       @parser.on("--open", 
@@ -672,8 +892,7 @@ EOD
 
       @parser.on("--[no-]viewer [VIEWER]", 
                  "Runs VIEWER to show the plot obtained --",
-                 "or doesn't " +
-                 "show up the viewer at all") do |x|
+                 "or doesn't show up the viewer at all") do |x|
         @viewer = x
       end
       
@@ -704,8 +923,8 @@ EOD
       
       @parser.on("--include FILE",
                  "Imports the file's functions into ctioga's",
-                 "namespace " +
-                 "so that they can be used", "by backends") do |file|
+                 "namespace so that they can be used", 
+                 "by backends") do |file|
         read_config_file(file)
       end
 
@@ -748,6 +967,23 @@ EOD
                  "The PDF output is kept in any case") do |v|
         @eps = v
       end
+
+      @parser.on("--png SIZE", 
+                 "Runs ctioga as usual, but runs convert afterwards",
+                 "to produce a nice PNG file") do |s|
+        @png = true
+        s =~ /(\d+)x(\d+)/
+        @png_size = [$1,$2]
+        # We setup the real size so that it matches the number of
+        # postscript points:
+        @real_size = "#{$1}bpx#{$2}bp"
+      end
+
+      @parser.on("--png-oversampling NB", 
+                 "How many more points to produce before downscaling") do |s|
+        @png_oversampling = s.to_f
+      end
+
       @parser.on("--args FILE",
                  "Reads argument from FILE, *one per",
                  "line* and then resumes normal processing") do |v|
@@ -757,6 +993,8 @@ EOD
       end
 
       @parser.on_tail("-h", "--help", "Show this message") do
+        puts banner
+        puts 
         puts @parser
         puts 
         puts "Please note that all the options can be abbreviated"
@@ -765,6 +1003,24 @@ EOD
       end
     end
 
+
+    # Start the next object in a complex plot (--grid, --col...)
+    def next_object
+        old = current_object
+        self.current_object = current_object.parent
+        a = @next_spec.dup      # We need to dup, else shift would make the
+        # next curve not that happy...
+        cls = a.shift
+        a <<  current_object
+        plot = cls.new(*a)
+        @current_object.layout.add_child(GridItemLayout.new(plot))
+        @current_object.add_elem(plot)
+        self.current_object = plot
+        
+        @next_block.call(old, plot) if @next_block
+    end
+
+
     def unshift_cmdline_args(*args)
       @args.unshift(*args)
     end
@@ -807,17 +1063,17 @@ EOD
       end
         
       # Scale plots if necessary:
-      data.mul!(:x,@x_factor) if @x_factor
-      data.mul!(:y,@y_factor) if @y_factor
+      data.mul!(:x,x_factor) if x_factor
+      data.mul!(:y,y_factor) if y_factor
 
       # Implement a logarithmic scale
-      data.safe_log10!(:x) if @x_log
-      data.safe_log10!(:y) if @y_log
+      data.safe_log10!(:x) if x_log
+      data.safe_log10!(:y) if y_log
 
       data.strip_nan
 
       # Then we create a curve object and give it style
-      if @histogram
+      if histogram
         curve = Histogram2D.new(get_current_style(set)) 
       else
         curve = Curve2D.new(get_current_style(set)) 
@@ -832,24 +1088,18 @@ EOD
       @last_curve = curve       # And we update the last_curve pointer.
     end
 
-    def frames_str(t)
-      return "top : #{t.frame_top} bottom: #{t.frame_bottom} " +
-        "left : #{t.frame_left} right: #{t.frame_right}"
-    end
-
-    
     # Provides the name of an output file, given its suffix and
     # optionnaly it base name (defaults to @figure_name).
     def output_filename(suffix, base = @fig_name) 
       if @figmaker.save_dir
-        File.join(t.save_dir,"#{base}#{suffix}")
+        File.join(@figmaker.save_dir,"#{base}#{suffix}")
       else
         "#{base}#{suffix}"
       end
     end
 
-    # Runs the plots, using the arguments given in the command-line
-    def run(args)
+    # Parses _args_ as command-line arguments.
+    def parse_command_line_arguments(args)
 
       # A little trick to allow in-place modification of the
       # command-line from an option.
@@ -864,14 +1114,6 @@ EOD
 
       @command_line = "#{File.basename($0)} #{quoted_args}"
 
-      puts <<"EOBANNER"
-This is ctioga version #{CTioga.version}, copyright (C) 2006 Vincent Fourmond
-ctioga comes with absolutely NO WARRANTY. This is free software, you are 
-welcome to redistribute it under certain conditions. See the COPYING file
-in the original tarball for details. You are also welcome to contribute if
-you like it, see in the manual page where to ask.
-EOBANNER
-
       # Read the CTIOGA environment variable 
       if ENV.key? "CTIOGA"
         ctioga_env = Shellwords.shellwords(ENV["CTIOGA"])
@@ -882,24 +1124,31 @@ EOBANNER
         send :ctioga_defaults
       end
 
-      @parser.order!(@args) do |spec|
-        # We use Backend#expand to leave room for the backend to
-        # interpret the text as many different sets.
-        sets = expand_spec(spec)
-        info "Expanding #{spec} to #{sets.join(', ')}"
-        for set in sets
-          process_data_set(set)
+      first = true
+      begin
+        @parser.order!(@args) do |spec|
+          # We use Backend#expand to leave room for the backend to
+          # interpret the text as many different sets.
+          sets = expand_spec(spec)
+          info "Expanding #{spec} to #{sets.join(', ')}"
+          # Auto --next: 
+          next_object if (@auto_next == :element and !first)
+          for set in sets
+            next_object if (@auto_next == :dataset and !first)
+            process_data_set(set)
+            first = false
+          end
         end
+      rescue OptionParser::InvalidOption => o
+        error "Invalid option: #{o.args.first}. Please run " + 
+          "ctioga --help to have a list of valid ones"
+        exit
       end
-      
-      # now, we plot the stuffs... if there is actually
-      # something to be plotted ;-) !
-
-      t = FigureMaker.new
-      @figmaker = t             # Boaf...
+    end
 
+    # Setups up various figuremaker variables
+    def setup_figure_maker(t)
       # Setting the TeX fontsize if default is not adequate
-
       t.tex_fontsize = @tex_fontsize if @tex_fontsize
 
       # add hyperref preamble to have nice PDF comments:
@@ -912,14 +1161,6 @@ EOBANNER
           ")\n}\n"
       end
 
-      if @debug
-        debug_figmaker(t)
-      end
-
-      # Now, we must use real_size stuff !! (default 12cmx12cm)
-      info "Producing PDF file of real size #{@real_size}"
-      setup_real_size(t) 
-
       if @debug_patterns
         debug_patterns(t)
       end
@@ -932,10 +1173,10 @@ EOBANNER
       # We put the command line used in the .tex file:
       t.tex_preview_preamble += 
         "% Produced by ctioga, with the command-line\n" +
-        "% #{@command_line}\n"
+        "% #{@command_line.gsub(/\n/,"\n% ")}\n"
       
-      if @decimal_separator
-        t.tex_preamble += <<'EOD'.gsub(',',@decimal_separator) # Neat !
+      if decimal_separator
+        t.tex_preamble += <<'EOD'.gsub(',',decimal_separator) # Neat !
 \def\frenchsep#1{\frenchsepma#1\endoffrenchsep} 
 \def\fseat#1{\frenchsepma}
 \def\frenchsepma{\futurelet\next\frenchsepmw}
@@ -953,14 +1194,6 @@ EOD
       # Add custom preamble:
       t.tex_preview_preamble += @preamble
 
-      
-
-      # Cleanup code
-      # off by default, as this could cause problems for
-      # legend pictures generation/EPS generation.
-      # The files will get cleaned up anyway.
-      t.autocleanup = false
-
       # If the initialization function was found, we call it
       if respond_to? :ctioga_init
         send :ctioga_init, t
@@ -969,16 +1202,86 @@ EOD
       for funcall in @init_funcalls
         funcall.do(t)
       end
+    end
+
+    # Runs the code to make the actual figure.
+    def do_figure(t)
+      @root_object.do(t)
+    end
+
+    # Ctioga's banner
+    def banner
+      <<"EOBANNER" 
+This is ctioga version #{CTioga.version}, copyright (C) 2006, 2007 by
+Vincent Fourmond ctioga comes with absolutely NO WARRANTY. This is
+free software, you are welcome to redistribute it under certain
+conditions. See the COPYING file in the original tarball for
+details. You are also welcome to contribute if you like it, see in
+the manual page where to ask.   
+EOBANNER
+    end
+
+    # Runs the plots, using the arguments given in the command-line.
+    # Now delegates to some other small functions.
+    def run(args)
+
+
+      # First, we parse command-line arguments:
+      parse_command_line_arguments(args)
+      
+      # now, we plot the stuffs... if there is actually
+      # something to be plotted ;-) !
+
+      @figmaker = t = FigureMaker.new
+      # Cleanup code off by default, as this could cause problems for
+      # legend pictures generation/EPS generation.
+      # The files will get cleaned up anyway.
+      t.autocleanup = false
+
+      # Now, we must use real_size stuff !! (default 12cmx12cm)
+      # This is not part of setup_figure_maker, as it can strongly interfere
+      # when ctioga is used to produce a figure.
+      info "Producing PDF file of real size #{@real_size}"
+      setup_real_size(t) 
+
+      # And, after that, the generic setup of the FigureMaker object.
+      setup_figure_maker(t)
+      
 
       t.def_figure(@fig_name) {
-        @root_object.do(t)
+        do_figure(t)
       }
 
       debug "Contents of the main FigureMaker object: " +
         figmaker_options(t)
 
       info "Generating file '#{@fig_name}.pdf'"
-      t.make_preview_pdf(t.figure_index(@fig_name))
+      if @fast
+        # It isn't that fast, actually.
+        info "Producing a fast version"
+        # We output only the temporary PDF file
+        class << t
+          undef show_text
+          # We replace show_text by show_marker, to at least get text
+          # Doesn't work for axes and ticks... Pity ! 
+          def show_text(dict)
+            for key in %w{side shift position pos justification}
+              dict.delete key
+            end
+            show_marker(dict)
+          end
+        end
+        t.create_figure_temp_files(t.figure_index(@fig_name))
+        base_name = if t.save_dir
+                      File.join(t.save_dir,@fig_name)
+                    else
+                      @fig_name
+                    end
+        # Remove the _figure.txt files
+        File.rename(base_name + "_figure.pdf", base_name + ".pdf")
+      else
+        t.make_preview_pdf(t.figure_index(@fig_name))
+      end
 
       if @eps
         info "Attempting to create an EPS file, watch out error messages"
@@ -999,16 +1302,18 @@ EOD
         orig.close
         out.close
 
-        cmd = "latex #{new}"
-        debug "Running #{cmd}"
-        system cmd
-
-        cmd = "dvips -o #{output_filename('.eps')} -t unknown -T " +
+        spawn "latex #{new}"
+        spawn "dvips -o #{output_filename('.eps')} -t unknown -T " +
           "#{@real_size.gsub(/x/,',')} #{output_filename('.new.dvi')}"
-        debug "Running #{cmd}"
-        system cmd
-        
       end
+
+      if @png
+        # We create the PNG file:
+        spawn "convert -density #{(@png_oversampling * 72).to_i} "+
+          "#{output_filename('.pdf')} -resize " +
+          "#{@png_size.join('x')} #{output_filename('.png')}"
+      end
+
       
       # Generating the separate legends if requested
       if not @separate_legend_styles.empty?
@@ -1040,7 +1345,7 @@ EOD
           begin
             File.rename(base_name + "_figure.pdf", base_name + ".pdf")
             if @eps             # Create also EPS files:
-              system "pdftops -eps #{base_name}.pdf"
+              spawn "pdftops -eps #{base_name}.pdf"
             end
             File.delete(base_name + ".tex")
             File.delete(base_name + "_figure.txt")
@@ -1078,12 +1383,13 @@ EOD
         files_to_remove << "pdflatex.log" 
         info "Cleaning up temporary files"
         debug "Cleaning #{files_to_remove.join ' '}"
-        files_to_remove.each {|f| begin
-                                    File.delete(f)
-                                  rescue
-                                    debug "File absent #{f}, not deleted" 
-                                  end
-        }
+        files_to_remove.each do |f| 
+          begin
+            File.delete(f)
+          rescue
+            debug "File #{f} absent, not deleted" 
+          end
+        end
       end
     end
 
@@ -1104,6 +1410,32 @@ EOD
       return str
     end
   end
+
+  # This function provides a 'command-line' interface from within Tioga !
+  def CTioga.define_tioga_figure(t, name, *args)
+    t.def_figure(name) {
+      CTioga.show_tioga_plot(t, *args)
+    }
+  end
+
+  # This function provides a 'command-line' interface from within Tioga !
+  def CTioga.show_tioga_plot(t, *args)
+    plot = PlotMaker.new
+    plot.parse_command_line_arguments(args)
+    plot.setup_figure_maker(t)
+    
+    width = 72 * t.
+      convert_output_to_inches(t.
+                               convert_figure_to_output_dx(t.convert_frame_to_figure_dx(1)))
+    height = 72 * t.
+      convert_output_to_inches(t.
+                               convert_figure_to_output_dy(t.convert_frame_to_figure_dy(1)))
+
+    plot.set_root_frame(0, width, height, 0)
+    
+    plot.do_figure(t)
+  end
+
   
 end
 
